package mediationci;

import java.util.Arrays;
import java.util.concurrent.*;
import java.util.logging.Level;
import java.util.logging.Logger;
import umontreal.iro.lecuyer.gof.KernelDensity;
import umontreal.iro.lecuyer.probdist.EmpiricalDist;
import umontreal.iro.lecuyer.probdist.NormalDist;
import umontreal.iro.lecuyer.randvar.ChiSquareGen;
import umontreal.iro.lecuyer.randvar.NormalACRGen;
import umontreal.iro.lecuyer.randvarmulti.MultinormalPCAGen;
import umontreal.iro.lecuyer.rng.BasicRandomStreamFactory;
import umontreal.iro.lecuyer.rng.LFSR113;
import umontreal.iro.lecuyer.rng.RandomStream;

/**
 *
 * @author falk
 * 
 * Copyright 2012-2014 Carl F. Falk
 * 
 * This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
public class HBCalc {

    private double lower; // main output variable of interest
    private double upper; // main output variable of interest
    private double path1;
    private double path2;
    private double se1;
    private double se2;
    private double alphaL;
    private double alphaU;
    private int df1;
    private int df2;
    private int rep;
    private int nDraws;
    private BasicRandomStreamFactory rstreamf; // random number gen
    private NormalACRGen normalGen; // random number gen
    private NormalDist nDist; // for density estimation
    private double mu[];
    private double sigma[][];
    private int progress;
    boolean done;
    private String calcMethod;  

    // constructors
    public HBCalc(double path1, double path2, double se1, double se2, int df1, int df2,
            int rep, int nDraws, double pci) {
        this.path1 = path1;
        this.path2 = path2;
        this.se1 = se1;
        this.se2 = se2;
        this.alphaL = (1-pci)/2;
        this.alphaU = 1-alphaL;
        this.df1 = df1;
        this.df2 = df2;
        this.rep = rep;
        this.nDraws = nDraws;
        //this.rstream = new LFSR113();
        this.rstreamf = new BasicRandomStreamFactory(LFSR113.class);
        this.normalGen = new NormalACRGen(new LFSR113(), new NormalDist());
        this.nDist = new NormalDist();
        this.progress = 0;
        this.done = false;
        this.calcMethod = "HB";
    }

    public HBCalc(double path1, double path2, double se1, double se2, int df1, int df2) {
        this.path1 = path1;
        this.path2 = path2;
        this.se1 = se1;
        this.se2 = se2;
        this.alphaL = .025;
        this.alphaU = 1-alphaL;
        this.df1 = df1;
        this.df2 = df2;
        this.rep = 300;
        this.nDraws = 100000;
        this.normalGen = new NormalACRGen(new LFSR113(), new NormalDist());
        //this.rstream = new LFSR113();
        this.rstreamf = new BasicRandomStreamFactory(LFSR113.class);        
        this.nDist = new NormalDist();
        this.progress = 0;
        this.done = false;
        this.calcMethod = "HB";        
    }

    public HBCalc(double path1, double path2, double se1, double se2, double cor,
            int rep, int nDraws, double pci) {
        this.path1 = path1;
        this.path2 = path2;
        this.se1 = se1;
        this.se2 = se2;
        this.alphaL = (1-pci)/2;
        this.alphaU = 1-alphaL;
        this.rep = rep;
        this.nDraws = nDraws;
        //this.rstream = new LFSR113();
        this.rstreamf = new BasicRandomStreamFactory(LFSR113.class);
        this.normalGen = new NormalACRGen(new LFSR113(), new NormalDist());
        this.nDist = new NormalDist();
        this.mu = new double[2];
        this.mu[0] = path1;
        this.mu[1] = path2;
        this.sigma = new double[2][2];
        this.sigma[0][0]=se1*se1;
        this.sigma[0][1]=cor*(se1*se2);
        this.sigma[1][0]=cor*(se1*se2);
        this.sigma[1][1]=se2*se2;
        
        this.progress = 0;
        this.done = false;
        this.calcMethod = "MC";
    }    
    
    public HBCalc(double path1, double path2, double se1, double se2, double cor) {
        this.path1 = path1;
        this.path2 = path2;
        this.se1 = se1;
        this.se2 = se2;
        this.alphaL = .025;
        this.alphaU = 1-alphaL;
        this.rep = 300;
        this.nDraws = 100000;
        this.normalGen = new NormalACRGen(new LFSR113(), new NormalDist());
        //this.rstream = new LFSR113();
        this.rstreamf = new BasicRandomStreamFactory(LFSR113.class);        
        this.nDist = new NormalDist();
        
        this.mu = new double[2]; //
        this.mu[0] = path1;
        this.mu[1] = path2;
        this.sigma = new double[2][2];
        this.sigma[0][0]=se1*se1;
        this.sigma[0][1]=cor*(se1*se2);
        this.sigma[1][0]=cor*(se1*se2);
        this.sigma[1][1]=se2*se2;
        
        this.progress = 0;
        this.done = false;
        this.calcMethod = "MC";                
    }  

    // getters/setters
    public double getAlphaL(){
        return alphaL;
    }
    
    public void setAlphaL(double alphaL){
        this.alphaL = alphaL;
    }
    
    public double getAlphaU(){
        return alphaU;
    }
    
    public void setAlphaU(double alphaU){
        this.alphaU = alphaU;
    }
    
    public int getDf1() {
        return df1;
    }

    public void setDf1(int df1) {
        this.df1 = df1;
    }

    public int getDf2() {
        return df2;
    }

    public void setDf2(int df2) {
        this.df2 = df2;
    }

    public double getPath1() {
        return path1;
    }

    public void setPath1(double path1) {
        this.path1 = path1;
    }

    public double getPath2() {
        return path2;
    }

    public void setPath2(double path2) {
        this.path2 = path2;
    }
    
    public double getSE1() {
        return se1;
    }

    public void setSE1(double se1) {
        this.se1 = se1;
    }

    public double getSE2() {
        return se2;
    }

    public void setSE2(double se2) {
        this.se2 = se2;
    }
    
    public double getLower(){
        return lower;
    }

    public double getUpper(){
        return upper;
    }
    public int getnDraws() {
        return nDraws;
    }

    public void setnDraws(int nDraws) {
        this.nDraws = nDraws;
    }

    public int getrep(){
        return rep;
    }
    
    public void setrep(int rep){
        this.rep = rep;
    }   


    public int getProgress(){
        return progress;
    }

    public void setProgress(int progress){
        this.progress = progress;
    }

    public boolean getDone(){
        return done;
    }
    
    public void setCalcMethod(String calcmethod){
        this.calcMethod = calcmethod;
    }

    public String getCalcMethod(){
        return calcMethod;
    }    
    
    public void calcHBVal() {
        EmpiricalDist ecdf;
        setProgress(0);
        this.done = false;
        double qL[] = new double[rep]; // store quantile estimates
        double qU[] = new double[rep];
        double dL[] = new double[rep]; // store densities
        double dU[] = new double[rep];
        double countL; // used for counting number of condition that hold true
        double countU;
        double cn;
        double nn=1;
        double cn1;
        double[] tmp = null;
        
        //Initial Estimates
        // Get product of posteriors
        tmp = this.getPost();
//        if("HB".equals(this.calcMethod)){        
//            tmp = this.getPostHB();            
//        } else if ("MC".equals(this.calcMethod)){
//            tmp = this.getPostMC();            
//        }
        
        // get quantiles
        Arrays.sort(tmp);       
        qL[0]=tmp[(int)Math.round(alphaL*nDraws)-1];
        qU[0]=tmp[(int)Math.round(alphaU*nDraws)-1];  

        // Compute densities
        ecdf = new EmpiricalDist(tmp);
        
        double[] test1 = new double[1];
        test1[0]=qL[0];
        double[] test2 = new double[1];
        test2[0]=qU[0];
        dL[0]=KernelDensity.computeDensity(ecdf, nDist, test1)[0];  
        dU[0]=KernelDensity.computeDensity(ecdf, nDist, test2)[0];  
        
        // Stochastic approximation/update
        setProgress(getProgress()+1);

        for(int n=1; n<rep; n++){
            
            // Get product of posteriors
            tmp=this.getPost();
//            if("HB".equals(this.calcMethod)){
//                tmp = this.getPostHB();    
//            } else if ("MC".equals(this.calcMethod)){
//                tmp = this.getPostMC();
//            }

            // Update Density
            // Test number of times abs value of CID-lastqL <= something
            countU=0;
            countL=0;
            cn = (1/Math.sqrt(nn));
            cn1= (Math.sqrt(nn-1));
            for(int i=0; i<nDraws; i++){
                if(Math.abs(tmp[i]-qL[n-1] )<= cn){
                    countL++;
                }
                if(Math.abs(tmp[i]-qU[n-1] )<= cn){
                    countU++;
                }                
            }

            
            dL[n] = (1-(1/nn))*dL[n-1] + (1/nn)*(countL/(2*cn*nDraws));
            dU[n] = (1-(1/nn))*dU[n-1] + (1/nn)*(countU/(2*cn*nDraws));
            
            // Update Quantile
            countL=0;
            countU=0;
            for(int i=0; i<nDraws; i++){
                if(tmp[i]<=qL[n-1]){
                    countL++;
                }
                if(tmp[i]<=qU[n-1]){
                    countU++;
                }                
            }
            // R code implementation
            // qL[n] = qL[n-1] + (1/nn)*(1/Math.max(dL[0]*cn1,dL[n-1]))*(alphaL-(countL/nDraws));
            // qU[n] = qU[n-1] + (1/nn)*(1/Math.max(dU[0]*cn1,dU[n-1]))*(alphaU-(countU/nDraws));            
            
            // current implementation
            qL[n] = qL[n-1] + (1/nn)*(Math.min(1/dL[n-1],(1/dL[0])*cn1))*(alphaL-(countL/nDraws));
            qU[n] = qU[n-1] + (1/nn)*(Math.min(1/dU[n-1],(1/dU[0])*cn1))*(alphaU-(countU/nDraws));                        
            
            setProgress(getProgress()+1);
            nn++;
        }
              
        this.lower = qL[rep-1];
        this.upper = qU[rep-1];
        this.done = true;
    }

    // Now can utilize multiple cores
/*    
    private double[] getPostHB(){
      
        double[] post1 = new double[nDraws];
        double[] post2 = new double[nDraws];
        
        // get posterior for each reg coefficient

        // multicore processing
        int np = Runtime.getRuntime().availableProcessors();        
        ExecutorService multicore = Executors.newFixedThreadPool(np);
        int indx=0;
        Future[] farray = new Future[np+1];  
        Future[] p1array = new Future[np+1];
        int section = nDraws / np;
        int remainder = nDraws % np;        
        double[] tmpp1;
        double[] tmpp2;
        double[] tmpp;        
        Callable m = null;
        Future f;
        
       // most of the result for p1
        for(int i=0; i<(np); i++){
            m = new postHB(section,path1,se1,df1);
            f = multicore.submit(m);
            farray[i] = f;
            indx+=section;
        }
        // rest of the result
        m = new postHB(remainder,path1,se1,df1);
        f = multicore.submit(m);
        farray[np] = f;        
        // combine results
        indx=0;
        for(int i=0; i<np; i++){
            try {
                tmpp = (double[]) farray[i].get();
                System.arraycopy(tmpp, 0, post1, indx, section);
                indx+=section;
            } catch (InterruptedException ex) {
                Logger.getLogger(HBCalc.class.getName()).log(Level.SEVERE, null, ex);
            } catch (ExecutionException ex) {
                Logger.getLogger(HBCalc.class.getName()).log(Level.SEVERE, null, ex);
            }
            //indx+=section;            
        }
        try {
            tmpp = (double[]) farray[np].get();
            System.arraycopy(tmpp, 0, post1, indx, remainder);
            indx+=section;
        } catch (InterruptedException ex) {
            Logger.getLogger(HBCalc.class.getName()).log(Level.SEVERE, null, ex);
        } catch (ExecutionException ex) {
            Logger.getLogger(HBCalc.class.getName()).log(Level.SEVERE, null, ex);
        }
        
       // most of the result for p2
        for(int i=0; i<(np); i++){
            m = new postHB(section,path2,se2,df2);
            f = multicore.submit(m);
            farray[i] = f;
            indx+=section;
        }
        // rest of the result
        m = new postHB(remainder,path2,se2,df2);
        f = multicore.submit(m);
        farray[np] = f;        
        // combine results
        indx=0;
        for(int i=0; i<np; i++){
            try {
                tmpp = (double[]) farray[i].get();
                System.arraycopy(tmpp, 0, post2, indx, section);
                indx+=section;
            } catch (InterruptedException ex) {
                Logger.getLogger(HBCalc.class.getName()).log(Level.SEVERE, null, ex);
            } catch (ExecutionException ex) {
                Logger.getLogger(HBCalc.class.getName()).log(Level.SEVERE, null, ex);
            }
            //indx+=section;            
        }
        try {
            tmpp = (double[]) farray[np].get();
            System.arraycopy(tmpp, 0, post2, indx, remainder);
            indx+=section;
        } catch (InterruptedException ex) {
            Logger.getLogger(HBCalc.class.getName()).log(Level.SEVERE, null, ex);
        } catch (ExecutionException ex) {
            Logger.getLogger(HBCalc.class.getName()).log(Level.SEVERE, null, ex);
        }        

        double[] newCID = new double[nDraws];

        // most of the result
        indx=0;
        for(int i=0; i<np; i++){
            tmpp1 = new double[section];
            tmpp2 = new double[section];
            for(int j=0; j<section; j++){
                tmpp1[j] = post1[indx+j];
                tmpp2[j] = post2[indx+j];
            }
            
            m = new multiplyCID(tmpp1,tmpp2,section);
            f = multicore.submit(m);
            farray[i] = f;
            indx+=section;
        }
        // rest of the result
        tmpp1 = new double[remainder];
        tmpp2 = new double[remainder];     
        for(int j=0; j<remainder; j++){
            tmpp1[j] = post1[indx+j];
            tmpp2[j] = post2[indx+j];
        }      
        m = new multiplyCID(tmpp1,tmpp2,remainder);
        f = multicore.submit(m);
        farray[np] = f;
        
        // most of the result
        indx=0;
        for(int i=0; i<np; i++){
            try {
                tmpp = (double[]) farray[i].get();
                System.arraycopy(tmpp, 0, newCID, indx, section);
                indx+=section;
            } catch (InterruptedException ex) {
                Logger.getLogger(HBCalc.class.getName()).log(Level.SEVERE, null, ex);
            } catch (ExecutionException ex) {
                Logger.getLogger(HBCalc.class.getName()).log(Level.SEVERE, null, ex);
            }
            
        }
        // rest of the result
        try {
            tmpp = (double[]) farray[np].get();
            System.arraycopy(tmpp, 0, newCID, indx, remainder);
            //indx+=section;
        } catch (InterruptedException ex) {
            Logger.getLogger(HBCalc.class.getName()).log(Level.SEVERE, null, ex);
        } catch (ExecutionException ex) {
            Logger.getLogger(HBCalc.class.getName()).log(Level.SEVERE, null, ex);
        }    
        
        multicore.shutdown();
        return newCID;
    }
     */
    
    private double[] getPost(){
      
        //double[] post1 = new double[nDraws];
        //double[] post2 = new double[nDraws];
        double[] newCID = new double[nDraws];
        
        // get posterior for each reg coefficient

        // multicore processing
        int np = Runtime.getRuntime().availableProcessors();   
        ExecutorService multicore = Executors.newFixedThreadPool(np);
        int indx=0;
        Future[] farray = new Future[np+1];  
        Future[] p1array = new Future[np+1];
        int section = nDraws / np;
        int remainder = nDraws % np;        
        //double[] tmpp1;
        //double[] tmpp2;
        double[] tmpp;        
        Callable m = null;
        Future f;
        
       // most of the result for p1
        for(int i=0; i<(np); i++){
            if("MC".equals(this.calcMethod)){
                m = new postMC(section);
            } else if ("HB".equals(this.calcMethod)){
                m = new postHB(section);
            }
            f = multicore.submit(m);
            farray[i] = f;
            indx+=section;
        }
        // rest of the result
        if("MC".equals(this.calcMethod)){    
            m = new postMC(remainder);
        } else if ("HB".equals(this.calcMethod)){
                m = new postHB(section);            
        }
        f = multicore.submit(m);
        farray[np] = f;
        // combine results
        indx=0;
        for(int i=0; i<np; i++){
            try {
                tmpp = (double[]) farray[i].get();
                System.arraycopy(tmpp, 0, newCID, indx, section);
                indx+=section;
            } catch (InterruptedException ex) {
                Logger.getLogger(HBCalc.class.getName()).log(Level.SEVERE, null, ex);
            } catch (ExecutionException ex) {
                Logger.getLogger(HBCalc.class.getName()).log(Level.SEVERE, null, ex);
            }
            //indx+=section;            
        }
        try {
            tmpp = (double[]) farray[np].get();
            System.arraycopy(tmpp, 0, newCID, indx, remainder);
            indx+=section;
        } catch (InterruptedException ex) {
            Logger.getLogger(HBCalc.class.getName()).log(Level.SEVERE, null, ex);
        } catch (ExecutionException ex) {
            Logger.getLogger(HBCalc.class.getName()).log(Level.SEVERE, null, ex);
        }
        
        multicore.shutdown();
        return newCID;
    }    

//    private double getPosterior(double path, double se, int df, RandomStream r) {
//        double output;
//        double z, chi;
//        z = this.normalGen.nextDouble();
//        chi = ChiSquareGen.nextDouble(r, df);
//        output = (z*Math.sqrt(df)*se)/(Math.sqrt(chi)) + path;
//        return output;
//    }
    
    private double getPosteriorHB(RandomStream r) {
        double output;
        double z1, z2, chi1, chi2;
        z1 = this.normalGen.nextDouble();
        chi1 = ChiSquareGen.nextDouble(r, this.df1);
        output = (z1*Math.sqrt(this.df1)*this.se1)/(Math.sqrt(chi1)) + this.path1;
        z2 = this.normalGen.nextDouble();
        chi2 = ChiSquareGen.nextDouble(r, this.df2);
        output *= (z2*Math.sqrt(this.df2)*this.se2)/(Math.sqrt(chi2)) + this.path2;      
        return output;
    }
    
    private double getPosteriorMC(MultinormalPCAGen r) {
        double[] output = new double[2];
        r.nextPoint(output);
        //MultinormalPCAGen.nextPoint(new NormalACRGen(r,nDist), mu, sigma, output);
        return output[0]*output[1];
    }

//    private double[] getPosterior(int n, double path, double se, int df) {
//        double[] output = new double[n];
//        RandomStream r = rstreamf.newInstance();               
//        for (int i = 0; i < n; i++) {
//            output[i] = getPosterior(path, se, df, r);
//        }
//        return output;
//    }
    
    private double[] getPosteriorHB(int n) {
        double[] output = new double[n];
        RandomStream r = rstreamf.newInstance();  
        for (int i=0; i < n; i++) {
            output[i] = getPosteriorHB(r);
        }
        return output;
    }    
    
    private double[] getPosteriorMC(int n) {
        double[] output = new double[n];
        RandomStream r = rstreamf.newInstance();
        MultinormalPCAGen mvrnorm = new MultinormalPCAGen(new NormalACRGen(r,nDist),mu,sigma);
        for (int i = 0; i < n; i++) {
            output[i] = getPosteriorMC(mvrnorm);
        }
        return output;
    }    
    
    // below used for some basic multi-core processing to speed things up
//    class postHB implements Callable {
//        private double[] post;
//        private int nd;
//        private double p;
//        private double s;
//        private int d;
//        public postHB(int nd, double p, double s, int d){
//            this.nd=nd;
//            this.p=p;
//            this.s=s;
//            this.d=d;
//        }
        
//        @Override
//        public double[] call(){
//            post = getPosterior(nd,p,s,d);    
//            return(post);
//        }
//    }
    
    class postHB implements Callable {
        private double[] post;
        private int nd;
        public postHB(int nd){
            this.nd=nd;
        }
        
        @Override
        public double[] call(){
            post = getPosteriorHB(nd);    
            return(post);
        }
    }    
    
    
    class postMC implements Callable {
        private double[] post;
        private int nd;
        private double p;
        private double s;
        private int d;
        public postMC(int nd){
            this.nd=nd;
        }
        
        @Override
        public double[] call(){
            post = getPosteriorMC(nd);    
            return(post);
        }
    }
    
    
//    class multiplyCID implements Callable {
//        private double[] CID;
//        private double[] p1;
//        private double[] p2;
//        private int nd;
        
//        public multiplyCID(double[] p1, double[] p2, int nd){
//            this.p1=p1;
//            this.p2=p2;
//            this.nd=nd;
//        }
        
//        @Override
//        public double[] call(){
//            CID = new double[nd];
//            for(int i=0; i<nd; i++){
//                CID[i] = p1[i]*p2[i];
//            }      
//            return(CID);
//        }
//    }    
    
    
    
}